//----------------------------------------------------------------------------
//  Description :
//      error reporting functions implementation
//----------------------------------------------------------------------------
//  Changes:
//  $Log__$
//----------------------------------------------------------------------------

#include <stdio.h>
#include <iostream.h>
#include <stdlib.h>
#include <stdarg.h>

#ifdef DEBUG
   #ifdef BRANKO
      #include "debug.h"
   #endif
#endif

#include "fsbool.h"
#include "fsanzopt.h"
#include "fserror.h"
#include "fspproc.h"

// obsolete
void FSError( enum FSErrorCode code )
{
   fprintf( stdout, "ERR. CODE=%d\n",code );
   exit( code );
}

const unsigned FSFATALERROR = 1;

static char *fsErrorStringTable[] =
{
  /* EX_USER_BREAK */             "!User break" , //Source analysis is interrupted by the user
  /* EX_TOO_MANY_ERR_WARN */      "!Too many errors or warnings" , //Broj nadjenih gresaka i upozorenja je prevazisao zadate granice
  /* EF_ERROR_READING_s */        "@Error reading %s" , //Unable to read the source from <filename> after the file was opened
  /* EF_UNABLE_CREATE_s */        "@Unable to create %s" ,
  /* EF_UNABLE_LOAD_s */          "@Unable to save %s" , //Unable to open or save file <filename>. Destination disk might not have enough space, or the user have no privilege to write to it
  /* EF_UNABLE_OPEN_s */          "@Unable to load %s" ,
  /* EF_UNABLE_SAVE_s */          "@Unable to open %s" , //Unable to open or load file <filename>. File does not exist, or the user have no privilege to read it
  /* E_ALREADY_IN_SU_s */         "%s: Already in struct/union" , //A structure or enumeration already has an element with the same name
  /* E_CONFL_PUB_SYM_s */         "%s: Conflicting global symbol" , //The project contains several global symbols of the same name and different type
  /* E_UNDEF_SYM_s */             "%s: Undefined symbol" , //Nedosaje definicija za koriscenu labela ili za deklarisanu 'static' funkciju
  /* E_MULT_INIT_s */             "%s: Multiple initialization" , //simbol je vise puta inicijalizovan. Dozvoljena je samo jedna inicijalizacija
  /* E_MULT_DEF_s */              "%s: Multiple definition" , //simbol ima vise definicija u okviru jednog modula
  /* E_MULT_DECL_s */             "%s: Multiple declaration" , //prethodna deklaracija simbola je takva da njegovo ponovno deklarisanje nije dozvoljeno (npr. simbol je lokalna promenljiva ili formalni parametar funkcije)
  /* E_NO_IDENT_REDECL_s */       "%s: Non-identical redeclaration" , //ponovljena deklaracija simbola  u okviru jednog modula nije kompatibilna sa prethodnom deklaracijom
  /* E_TOO_MANY_TYPESPECS */      "Too many type specifiers" , //Simbol ne moze istovremeno biti deklarisan kao 'long' i 'short' ili 'signed' i 'unsigned', ili kao 'int' i 'float'
  /* E_TOO_MANY_STORAGESPECS */   "Too many type storage specifiers" , //Deklaracija sadrzi nekompatibilne storage specifiers, npr. 'static' i 'extern'
  /* E_INV_USAGE_OF_TYPEDEF_s */  "%s: Invalid usage of typedef" , //Typedef simbol je koriscen kao ime promenljive ili funkcije
  /* E_NOT_ALLOWED_STORAGE_s */   "%s: Invalid storage class" , //A storage attribute is not allowed u kontekstu tekuce deklaracije
  /* E_NOT_ALLOWED_STORAGE */     "Invalid storage class" , //Navedena je storage class koja je neprimenljiva na mestu na kome se nalazi deklaracija
  /* E_INIT_NOT_ALLOWED_DECL */   "Initialization is not allowed" , //Inicijalizacija nije dozvoljena za 'typedef', 'extern' deklaracije ili deklaracije funkcija
  /* E_UNDEF_SU_s */              "%s: Undefined struct/union" , //Nepotpuno definisana struct/union ne moze biti sadrzana od druge struct/union
  /* E_NOT_FUNC_s */              "%s: Not a function" , //Deklaracija koja prethodi telu funkcije ne predstavlja deklaraciju funkcije
  /* E_NOT_FUNC_PARAM_s */        "%s: Not a function parameter" , //Old style function definition of a function contains declaration of an identifier which is not in its parameter list
  /* E_UNKNOWN_ID_s */            "%s: Unknown identifier" , //Upotreba nedeklarisanog identifikatora
  /* E_NOT_SU_s */                "%s: Not a struct/union on left side of '.'" , //Argument sa leve strane '.' operatora mora imati tip struct/union
  /* E_UNTERM_STR_CONST */        "Unterminated string constant" , //A string constant is not terminated with a double quote character
  /* E_UNTERM_CHAR_CONST */       "Unterminated character constant" , //A character constant is not terminated with a single quote character
  /* E_EMPTY_CHAR_CONST */        "Empty character constant" , // A character constant mora sadrzati najmanje jedan znak
  /* E_CHAR_CONST_TOO_LONG */     "Character constant too long" , // An ordinary character constant mora sadrzati najvise jedan znak
  /* E_INV_NUM_CONST */           "Invalid numeric constant" , //Neispravan format brojne konstante
  /* E_PARSER_STACK_OVERFLOW */   "Parse stack overflow" , //The parse stack has overflowed. Source contains extremely complex statements or expressions.
  /* E_INV_DECL */                "Invalid declaration" , // Syntax error in a declaration
  /* E_INV_STRUCT_ELEM */         "Invalid struct/union element" , // Syntax error in a struct/union element declaration
  /* E_INV_PARAM_DECL */          "Invalid parameter declaration" , // Syntax error in a function parameter declaration
  /* E_INV_STMT */                "Invalid statement" , // Syntax error in a statement
  /* E_INV_ARG_LIST */            "Invalid argument list" , //Sintaksna greska u listi argumenata u okviru poziva funkcije
  /* E_INV_EXPR */                "Invalid expression" , //Syntax error in an expression
  /* E_NOT_SUE_PTR_s */           "%s: Not a pointer to a struct/union" , //Argument sa leve strane '.' operatora mora imati tip struct/union
  /* E_INV_FUNC_CALL_s */         "%s: Invalid function call" , //poziv funkcije je upotrebljen nad simbolom koji nije funkcija ili pokazivac na funkciju
  /* E_INV_INDIR_s */             "%s: Invalid indirection argument" , //simbol nad kojim se primenjuje operator indirekcije '*' mora biti pointer ili ime niza
  /* E_INV_ARRAY_ACCESS_s */      "%s: Invalid array access" , //simbol nad kojim se primenjuje array-access operator mora biti pointer ili ime niza
  /* E_INV_ADDR_OP_s */           "%s: Invalid '&' operand " , //The address-of operator '&' je upotrebljen nad operandom neodgovarajuce vrste, reecimo nad enum konstantom
  /* E_INV_ASSIGN_s */            "%s: Invalid left side of assignment" , //Leva strana operatora dodele vrednosti lili argument '++' ili '--' operatora nije lvalue
  /* E_PUB_SYM_MULT_DEFS_s */     "%s: Global symbol with multiple definitions" , //Pronadjeno je vise definicija za simbol sa extern linkage
  /* E_UNKN_STRUCT_MEMB_ss */     "%s: Unknown struct/union member %s" , //Ime sa desne strane '->' ili '.' operatora ne predstavlja ni jedan element struct/union
  /* E_PP_UNTERM_COMMENT_ST_u */  "Unterminated comment started on line %u" , //A comment started with '/*' doesn't have a closing delimiter '*/' u okviru jedne izvorne datoteke
  /* E_PP_UNTERM_COND_ST_u */     "Unterminated conditional directive started on line %u" , //Pri nailasku na kraj izvorne datoteke nisu sve #if, #ifdef ili #ifndef preprocesorske direktive uparene sa odgovarajucom #endif direktivom
  /* E_PP_MACRO_INCL_OVERFLOW */  "Macro/include stack overflow" , //The number of nested macro/include expansions has become too large. If traditional preprocessor is active, this may indicate recursive macro definitions
  /* E_PP_DEFINED_WITOUT_ID */    "Operator 'defined' needs an identifier" , //Neispravan format upotrebe 'defined' operatora unutar #if preprocesorske direktive
  /* E_PP_INV_INCL_FILENAME */    "Invalid filename format in #include directive" , //Ime datoteke navedeno unutar #include direktive mora (posle eventualne makroekspanzije) biti oblika <filename> ili "filename"
  /* E_PP_OPEN_INCL_FILE_s */     "Unable to open include file %s" , //Unable to open include file <filename>. File does not exist, or the user have no privilege to read it
  /* E_PP_MISP_HH */              "Misplaced '#' or '##' in a macro definition" , //The stringify operator '#' is not followed by a macro argument name, or the concatenation operator '##' is not placed between two tokens
  /* E_PP_INV_MACRO_NAME */       "Invalid macro name" , //The name of a macro must ba a valid identifier
  /* E_PP_INV_MACRO_ARG_LIST */   "Invalid macro argument list" , //The macro argument list is not a comma separated list of identifiers enclosed in braces, or a argument name is repeated in the list
  /* E_PP_INV_X_DIR_s */          "Invalid #%s directive" , //neispravan ostatak linije sa #ifdef, #ifndef ili #line direktivom. U prva dva slucaja to mora biti identifikator, za #line direktivu je to broj linije i opciono ime datoteke.
  /* E_PP_MISP_X_DIR_s */         "Misplaced #%s directive" , //Direktive #else, #endif ili  #elif moraju biti u biti u odgovarajucoj sekvenci of condititonal compilation directives, koje moraju zapoceti sa #if, #ifdef ili #ifndef
  /* E_PP_UNKNOWN_DIR */          "Unknown preprocessor directive" , //ime koje sledi iza '#' nije prepoznato kao validna preprocesorska direktiva
  /* E_PP_UNTERM_MACRO_u */       "Unterminated macro call started on line %u" , //Detektovan je kraj izvorne datoteke, a da lista of actual arguments of somme macro nije kompletirana
  /* E_PP_INV_NUM_MACRO_ARGS_s */ "Macro '%s' used with wrong number of arguments" , //Broj of actual macro arguments se razlikuje od broja odredjenog definicijom makroa
  /* E_PP_INV_MACRO_DEF_s */      "Invalid predefined macro '%s'" ,
  /* E_PP_UNEX_EOF */             "Unexpected EOF" , //Kraj izvornog teksta je detektovan a da tekuca deklaracija ili statement nije zavrsena
  /* E_PP_ERROR_DIR_s */          "#error %s" , //Greska generisana #error direktivom. Prikazuje se korisnicki odredjen tekst iz ostatka linije koja sadrzi direktivu
  /* E_DIV_ZERO */                "Division by zero" , //Deljenje nulom u izracunavanju izraza u okviru preprocessor directive
  /* W_ID_LONG */                 "-Identifier too long, truncated" , //Only the first 32 characters of a identifier are taken to be significant
  /* W_ID_NEVER_USED_s */         "-%s: Identifier never used" , //Symbol is declared, but not used
  /* W_MOD_SYM_IN_FUNC_ss */      "-%s: Module level symbol used only in function %s" , //Variable or constant is declared as static at module level, but is used in only one function. Its definition should be moved to that function.
  /* W_MISSING_PUB_DEF_s */       "-%s: Missing global definition" , //symbol with external linkage is declared as 'extern', but no definition could be found. Ova poruka se ne ispisuje za simbole deklarisane unutar datoteka ucitanih sa '#include <>' direktivom i za include datoteke koje sadrze '#pragma lib' direktivu.
  /* W_GLOB_ID_USED_ONCE_s */     "-%s: Global identifier used in only one module" , //Symbol has external linkage, but is used in only one module
  /* W_NO_IDENT_TYPESPEC_s */     "-%s: Non-identical type specification" , //Dve deklaracije simbola se razlikuju po upotrebi 'signed', 'unsigned', 'const' ili 'volatile' atributa
  /* W_UNDEF_SUE_s */             "-%s: Undefined struct/union/enum" , //u modulu nije data potpuna definicija za navedeni struct/union/enum
  /* W_FUNC_CALL_WITHOUT_DCL_s */ "-%s: Function call without declaration" , //Poziv funkcije koja prethodno nije deklarisana. Podrazumevana deklaracija je 'int ident();'
  /* W_INV_ESCAPE */              "-Invalid escape sequence" , //A string or character constants containa an invalid sequence after '\' character
  /* W_OLD_ANSI_FUNC_DECS_s */    "-%s: Mixed old style and ANSI function declarations" , //Istovremeno postoje ANSI i old style deklaracije iste funkcije
  /* W_PP_NO_IDENT_MCR_REDEF_s */ "-Non-identical redefinition of macro '%s'" , //This macro definition differs from the previous one
  /* W_PP_WARNING_DIR_s */        "-#warning %s" , //Upozorenje generisano #warning direktivom. Prikazuje se korisnicki odredjen tekst iz ostatka linije koja sadrzi direktivu
  /* W_ASM_STMT */                "-Symbol references in 'asm' statement are ignored" , //Upozorenje da je koriscena asm, _asm ili _asm naredba. DA-C ne prepoznaje upotrebe simbloa unutar asemblerskih naredbi ili blokova. Usage of '#ifndef __DA_C__' is suggested:
  /* W_NONPORT_DCL_s */           "-%s: Nonportable declaration" , //Symbol declaration is given without any specifier, 'int' is assumed. Some compilers don't support such declarations.
  /* W_SIZEOF_IN_PP */            "-'sizeof' value is unknown" ,
  /* W_ARGT_IN_PP */              "-'_argt$' value  is unknown" ,
  /* W_PP_ALREADY_INCL_s */       "-File %s recursively included"  // File <FILENAME> contains a #include directive which produces repeated inclusion of  <FILENAME>, or loads a header which includes it again.

};

#if defined( DEBUG ) || defined ( PPDEBUG )
static char ostr[ 200 ];
#endif

void FSAnzWarning( unsigned code, ... )
{
   va_list args;
   va_start( args, code );
#if defined( DEBUG ) || defined ( PPDEBUG )
   cout << "Warning " << preProc->File() << '(' << preProc->Line() << ") ";
   vsprintf( ostr, fsErrorStringTable[ code ], args );
   cout << ostr << '\n';
#else
   // ugraditi odgovarajuci kod
#endif
   va_end( args );
}

void FSAnzError( unsigned code, ... )
{
   va_list args;
   va_start( args, code );
#if defined( DEBUG ) || defined ( PPDEBUG )
   cout << "Error " << preProc->File() << '(' << preProc->Line() << ") ";
   vsprintf( ostr, fsErrorStringTable[ code ], args );
   cout << ostr << '\n';
#else
   // ugraditi odgovarajuci kod
#endif
   va_end( args );
}

void FSAnzFatalError( unsigned code, ... )
{
   va_list args;
   va_start( args, code );
#if defined( DEBUG ) || defined ( PPDEBUG )
   cout << "Fatal " << preProc->File() << '(' << preProc->Line() << ") ";
   vsprintf( ostr, fsErrorStringTable[ code ], args );
   cout << ostr << '\n';
#else
   // ugraditi odgovarajuci kod
#endif
   va_end( args );
   throw FSFATALERROR;
}

