//----------------------------------------------------------------------------
//  Description :
//      preprocessor file, macro & if stacks,
//      macro table implementation
//----------------------------------------------------------------------------
//  Changes:
//  $Log__$
//----------------------------------------------------------------------------

#include <string.h>
#include <fstream.h>
#include <stdlib.h>
#ifdef PPDEBUG
   #include <iostream.h>
#endif

#ifdef DEBUG
   #ifdef BRANKO
      #include "debug.h"
   #endif
#endif

#include "fsbool.h"
#include "fsanzopt.h"
#include "fserror.h"
#include "fspproc.h"
#include "fsstacks.h"
#include "fstoken.h"
//#include "fssymtab.h" // fspars_t.h needs it
#include "fspars_t.h"
//-------------------------------------------------------------------
//   class FSFile implementation
//-------------------------------------------------------------------
//##ModelId=419A04D203D2
FSFile::FSFile( FSPreProc &pProc, char *fileName )
	: FSFileStackElem( FS_STK_FILE ),
     preProc( pProc ),
	  lookAheadChar( ' ' ),
     savedIfStackBottom( preProc.ifStack.SetBottom() ),
	  line( 1 )
{
		name[ FS_MAX_FILE_NAME ] = '\0';
		strncpy( name, fileName, FS_MAX_FILE_NAME );
		file.open( name, ios::in );
		if ( file.fail() ) FSAnzError( E_PP_OPEN_INCL_FILE_s, name );
} // ugraditi append direktorijuma

//##ModelId=419A04D203BC
int FSFile::GetChar()
{
	int c;
   while( 1 )
	{
      c = lookAheadChar;
      lookAheadChar = file.get();
		if ( c != '\\' || lookAheadChar != '\n' ) break;
      else
      {
         line++;
         lookAheadChar = file.get();
      }
   }
   if ( c == '\n' ) line++;
   // make sure that there is always \n before EOF
   if ( lookAheadChar == EOF && c != '\n' && c != EOF ) lookAheadChar = '\n';
	return c;
}

//##ModelId=419A04D203E4
FSFile::~FSFile()
{
   preProc.ifStack.SetBottom( savedIfStackBottom );
}

//-------------------------------------------------------------------
//   class FSFileStack implementation
//-------------------------------------------------------------------
//##ModelId=419A04D300E3
void FSFileStack::PushParam( char *str )
{
#ifdef PPDEBUG
   cout << preProc.File() << '(' << preProc.Line() << ") ";
   cout << "FS PUSH PARAM top " << ( top + 1 ) << '\n';
#endif
	if ( top < FS_MAX_INCLUDE_LEVEL )
	{
		FSParam *pNewParam = new FSParam( str );
		stack[ top++ ] = pNewParam;
		pNewParam->savedTokenizerLookahead = preProc.tokenizer.nextChar;
		preProc.tokenizer.nextChar = ' ';
	}
	else FSAnzError( E_PP_MACRO_INCL_OVERFLOW );
}

//##ModelId=419A04D300ED
void FSFileStack::PushFile( char *fileName )
{
#ifdef PPDEBUG
   cout << preProc.File() << '(' << preProc.Line() << ") ";
   cout << "FS PUSH FILE top " << ( top + 1 ) << '\n';
#endif
	if ( top < FS_MAX_INCLUDE_LEVEL )
	{
		FSFile *pNewFile = new FSFile( preProc, fileName );
		if ( pNewFile->Status() != FSFile::SUCCESS )
		{
			delete pNewFile;
		}
		else
		{
			stack[ top++ ] = pNewFile;
			pNewFile->savedTokenizerLookahead = preProc.tokenizer.nextChar;
			preProc.tokenizer.nextChar = ' ';
		}
	}
	else FSAnzError( E_PP_MACRO_INCL_OVERFLOW );
}


//##ModelId=419A04D30101
void FSFileStack::Pop()
{
#ifdef PPDEBUG
   cout << preProc.File() << '(' << preProc.Line() << ") ";
   cout << "FS POP top " << ( top - 1 ) << '\n';
#endif
	preProc.tokenizer.nextChar = stack[ top - 1 ]->savedTokenizerLookahead;
	delete stack[ --top ];
}


//##ModelId=419A04D30117
FSFileStack::~FSFileStack()
{
   bottom = 0;
   while ( IsNotEmpty() )
         Pop();
}

//-------------------------------------------------------------------
//   class FSMacroArgs implementation
//-------------------------------------------------------------------

//##ModelId=419A04D3018D
int FSMacroArgs::LookUp( char *argName )
{
   int index;
   for ( index = number; index > 0; index-- )
   {
       if ( strcmp( argName, args[ index - 1 ] ) == 0 ) break;
   }
   return index;
}

//##ModelId=419A04D301DD
void FSMacroArgs::Insert( char *argName )
{
      // assumes there is room for one more arg
      if ( LookUp( argName ) == 0 )
      {
         args[ number++ ] = strcpy( new char[ strlen( argName ) + 1 ], argName );
      }
      else
      {
          FSAnzError( E_PP_INV_MACRO_ARG_LIST );
      }
}


//##ModelId=419A04D301A1
FSMacroArgs::FSMacroArgs( FSPreProc &preProc, int &token ) // return last token read while parsing the arg. list
{
	if ( preProc.tokenizer.nextChar != '(' )
   {
      // no arguments
      token = preProc.tokenizer.Get(); // set the first token for macro body parsing
      parenth = FALSE;
      number = 0;
      for ( int i = 0; i < FS_MAX_ARGS; i++ )
      {
          args[ i ] = 0;
      }
      return;
	}
   // macro with arguments
   parenth = TRUE;
   number = 0;
   token = preProc.tokenizer.Get(); // skipping LP
   enum Status { 
	   EXPECT_NAME, 
	   EXPECT_SEPARATOR, 
	   END_LIST 
   } status = EXPECT_NAME;
   while ( status != END_LIST )
   {
      switch ( token = preProc.tokenizer.Get() )
      {
         case NAME:
              if ( status != EXPECT_NAME )
              {
						FSAnzError( E_PP_INV_MACRO_ARG_LIST ); // assume missing comma
              }
              status = EXPECT_SEPARATOR;
              if ( number >= FS_MAX_ARGS )
              {
                  FSAnzError( E_PP_INV_MACRO_ARG_LIST );
                  // ignore the rest of list
                  int token;
                  while (    ( token = preProc.tokenizer.Get() ) != RP
                          && token != NEWLINE
                        )
                        ;
                  status = END_LIST;
				  }
              else
              {
                  Insert( preProc.outBuffer.GetTxt() );
              }
              break;

         case COMMA:
              if ( status != EXPECT_SEPARATOR )
              {
                  FSAnzError( E_PP_INV_MACRO_ARG_LIST );
              }
              status = EXPECT_NAME;
				  break;

         case RP:
              status = END_LIST;
              token = preProc.tokenizer.Get(); // sets first token for body parsing
              break;

         case NEWLINE:
              FSAnzError( E_PP_INV_MACRO_ARG_LIST );
              status = END_LIST;
              break;

         default:
				 FSAnzError( E_PP_INV_MACRO_ARG_LIST );
             status = END_LIST;
      }
   }
}

//##ModelId=419A04D301AC
FSMacroArgs::FSMacroArgs() // predefined macro with no args
{
      // no arguments
      parenth = FALSE;
      number = 0;
      for ( int i = 0; i < FS_MAX_ARGS; i++ )
      {
          args[ i ] = 0;
      }
}

//##ModelId=419A04D301B5
FSMacroArgs::~FSMacroArgs()
{
   for ( int i = 0; i < number; i++ )
       delete [] args[ i ];
}

//-------------------------------------------------------------------
//   class FSMacroBody implementation
//-------------------------------------------------------------------
//##ModelId=419A04D302FF
void FSMacroBody::FSMacroBodyElem::Insert( FSMacroBodyElem *& current, char *str, int len )
{
   while ( len > 0 )
   {
       int leftChars = current->txt + FS_MAX_MBODY_CHUNK_LEN - current->txtEnd;
       if ( len <= leftChars )
       {
          strncpy( current->txtEnd, str, len );
			 * ( current->txtEnd += len ) = '\0';
          break;
       }
		 else
       {
          strncpy( current->txtEnd, str, leftChars );
          * ( current->txtEnd += leftChars ) = '\0';
          str += leftChars;
          len -= leftChars;
          current->next = new FSMacroBodyElem;
          current = current->next;
       }
   }
}
//##ModelId=419A04D3031D
void FSMacroBody::FSMacroBodyElem::Insert( FSMacroBodyElem *&current, int arg, FSArgumentMode mode )
{
	current->arg = arg;
   current->argMode = mode;
   current->next = new FSMacroBodyElem;
   current = current->next;
}

//##ModelId=419A04D3027D
FSMacroBody::FSMacroBody( FSPreProc &preProc, FSMacroArgs &args, int &lastTokenRead )
{
	head = current = new FSMacroBodyElem;
	int prevToken = RP;
	int arg;
	while ( 1 )
	{
		switch ( lastTokenRead )
		{
			case SHARP:
					if ( prevToken != SHARPSHARP ) FSMacroBodyElem::Insert( current, " ", 1 );
               // kompajler se ovde zbuni oko upotrebe prevTokena mada je cela stvar u while petlji
					prevToken = lastTokenRead;
					lastTokenRead = preProc.tokenizer.Get();
					if ( lastTokenRead == NAME && ( arg = args.LookUp( preProc.outBuffer.GetTxt() )  ) != 0 )
					{
						// OK - arg follows #
						FSMacroBody::FSMacroBodyElem::Insert( current, "\"", 1 );
						FSMacroBody::FSMacroBodyElem::Insert( current, arg, FS_AR_STRING );
						FSMacroBody::FSMacroBodyElem::Insert( current, "\"", 1 );
					}
					// in all other cases # and the next token are copied in the body literally
					else
					{
						FSMacroBodyElem::Insert( current, "#", 1 );
						FSMacroBodyElem::Insert( current, preProc.outBuffer.GetTxt(), preProc.outBuffer.GetLen() );
					}
					prevToken = lastTokenRead;
					lastTokenRead = preProc.tokenizer.Get();
					break;

			case SHARPSHARP:
					prevToken = lastTokenRead;
					lastTokenRead = preProc.tokenizer.Get();
					break;

			case NAME:
					if ( prevToken != SHARPSHARP ) FSMacroBodyElem::Insert( current, " ", 1 );
					arg = args.LookUp( preProc.outBuffer.GetTxt() );
					if ( arg == 0 )
					{
						FSMacroBody::FSMacroBodyElem::Insert( current, preProc.outBuffer.GetTxt(), preProc.outBuffer.GetLen() );
						prevToken = lastTokenRead;
						lastTokenRead = preProc.tokenizer.Get();
					}
               else
               {
						int prevPrevToken = prevToken;
						prevToken = lastTokenRead;
						lastTokenRead = preProc.tokenizer.Get();
						// if ## precedes or follows the argument, it's not expanded
						FSMacroBodyElem::Insert( current, arg, ( prevPrevToken == SHARPSHARP || lastTokenRead == SHARPSHARP ) ? FS_AR_NORMAL : FS_AR_EXPAND );
					}
					break;

         case NEWLINE:
               goto whileEnd;

         default:
               if ( prevToken != SHARPSHARP ) FSMacroBodyElem::Insert( current, " ", 1 );
               FSMacroBody::FSMacroBodyElem::Insert( current, preProc.outBuffer.GetTxt(), preProc.outBuffer.GetLen() );
               prevToken = lastTokenRead;
               lastTokenRead = preProc.tokenizer.Get();
      }
   }
   whileEnd: ;
}

//##ModelId=419A04D30289
FSMacroBody::FSMacroBody()
{
	head = current = new FSMacroBodyElem;
}

//##ModelId=419A04D30291
FSMacroBody::~FSMacroBody()
{
   FSMacroBodyElem *prev;
   for ( current = head; current; )
   {
      prev = current;
      current = current->next;
      delete prev;
   }
}

//-------------------------------------------------------------------
//   class FSMacro implementation
//-------------------------------------------------------------------

//-------------------------------------------------------------------
//   class FSMacroTable implementation
//-------------------------------------------------------------------

//##ModelId=419A04D400A0
FSMacroTable::FSMacroTable( FSPreProc &pProc )
   : preProc( pProc )
{
   for ( int i = 0; i <= FS_HASH_DIM; i++ ) table[ i ] = 0;
}

//##ModelId=419A04D400A9
FSMacroTable::~FSMacroTable()
{
   FSMacro *current, *prev;
   for ( int i = 0; i <= FS_HASH_DIM; i++)
   {
      current = table[ i ];
      while ( current )
      {
         prev = current;
         current = current->Next();
         delete prev;
		}
	}
}

//##ModelId=419A04D400C6
unsigned FSMacroTable::HashValue( char *name )
{
	unsigned hv = 0;
	while ( *name )
			hv += *(name++);
	return hv % FS_HASH_DIM;
}

// returns a pointer to pointer to macro 'name' if it is present in the table, or a
// pointer to 'next' field in the last elem of the hash list where the
// macro should be placed
//##ModelId=419A04D400DA
FSMacro **FSMacroTable::InternLookUp( char *name )
{
	FSMacro **ppMacro = table + HashValue( name );
	while ( *ppMacro )
	{
		if ( strcmp( name, ( *ppMacro )->Name() ) == 0 )
      {
         break;
      }
		ppMacro  = &( ( *ppMacro )->Next() );
	}
	return ppMacro;
}

//-------------------------------------------------------------------
//   class FSMacroCallParams implementation
//-------------------------------------------------------------------

//##ModelId=419A04D401D4
void FSMacroCallParams::ProcessParamChunk( int &numParams, Status &status )
{
	if ( status == EXPECT_ARG && numParams < macro.Args().Number() )
	{
		nonExpParams [ ++numParams ] = InsertFirstChunk( preProc.outBuffer.GetTxt(), preProc.outBuffer.GetLen() );
			  // rem: nonExpParams[0] always ""
	}
	else if( status == EXPECT_ARG_OR_SEPARATOR && numParams <= macro.Args().Number() )
	{
		InsertNextChunkSpc( preProc.outBuffer.GetTxt(), preProc.outBuffer.GetLen() );
	}
	else
	{
	  // number of actuals > no. of formals; error reported elsewhere
     numParams++;
	}
	status = EXPECT_ARG_OR_SEPARATOR;
}

//##ModelId=419A04D4017A
FSMacroCallParams::FSMacroCallParams( FSMacro &mcro, FSPreProc &pProc )
	: macro( mcro ),
     preProc( pProc ),
     paramSpcFree( paramSpace )
{
	for ( int i = 0; i <= FS_MAX_ARGS; i++ )
	{
      nonExpParams[ i ] = "";
      expParams[ i ] = "";
   }
// in case macro has parenthesis, last token read is LP; in case macro has no parenthesis,
// last token read is not important
	if ( macro.Args().Parenth() == FALSE )
   {
		return;
   }
   // now collect actual parameters until RP or EOF
   Status status = EXPECT_ARG;
	int openParenth = 1;
	int numParams = 0;
	while ( status != END_LIST )
	{
		switch( preProc.tokenizer.Get() )
		{
			case COMMA:
					  if ( openParenth == 1 )
					  {
							status = EXPECT_ARG;
					  }
					  else
					  {
							ProcessParamChunk( numParams, status );
					  }
					  break;

			case EOF:
					  if ( preProc.macroCallStack.IsNotEmpty() )
					  {
							preProc.macroCallStack.Pop();
					  }
					  else status = END_LIST;
					  break;

			case LP:
					  openParenth++;
					  ProcessParamChunk( numParams, status );
					  break;

         case RP:
                 if ( --openParenth == 0 )
                 {
                     status = END_LIST;
                 }
					  else
					  {
                     ProcessParamChunk( numParams, status );
                 }
					  break;

         default:
                 ProcessParamChunk( numParams, status );
      }
	}
   int numArgs = macro.Args().Number();
	if ( numParams != numArgs )
   {
      FSAnzError( E_PP_INV_NUM_MACRO_ARGS_s, macro.Name() );
      if ( numParams > numArgs ) numParams = numArgs;
   }
	// now calculate expanded parameters
	// 'clear' stacks
	int savedMCallStkBottom = preProc.macroCallStack.SetBottom();
	int savedFilStkBottom = preProc.fileStack.SetBottom();
	for ( int parmIndex = 1; parmIndex <= numParams; parmIndex++ )
	{
		preProc.fileStack.PushParam( nonExpParams[ parmIndex ] );
		FSMacro *pMacro;
		InsertFirstChunk( "" , 0 );
		Boolean firstChunk = TRUE;
		int token = preProc.tokenizer.Get();
		while ( 1 )
		{
			switch ( token )
			{
				case EOF:
					if ( preProc.macroCallStack.IsNotEmpty() )
					{
						preProc.macroCallStack.Pop();
					}
					else
					{
						goto WhileEnd;
					}
					break;

				case NAME:
					if (     ( pMacro = preProc.macroTable.LookUp( preProc.outBuffer.GetTxt() ) ) != 0
							&& preProc.macroCallStack.ExpandingNow( pMacro ) == FALSE
						)
					{
						// there is a macro with that name, and this is not a recursive call
						if ( pMacro->Args().Parenth() == FALSE )  // macro need no parenthesis
						{
							preProc.macroCallStack.Push( *pMacro );
						}
						else if ( ( token = preProc.tokenizer.Get() ) == LP )// macro with parenthesis & there are parenth. following a name
						{
							preProc.macroCallStack.Push( *pMacro );
						}
						// not a macro call because macro needs '(' and there is no '('
						else if ( firstChunk )
						{
							expParams[ parmIndex ] = InsertFirstChunk( pMacro->Name(), strlen( pMacro->Name() ) );
							firstChunk = FALSE;
							continue; // in order to reexamine lookahead token
						}
						else
						{
							InsertNextChunkSpc( pMacro->Name(), strlen( pMacro->Name() ) );
							continue; // in order to reexamine lookahead token
						}
					}
					// token is not a macro name, or macro recursion
					else
					{
						if ( firstChunk )
						{
							 expParams[ parmIndex ] = InsertFirstChunk( preProc.outBuffer.GetTxt(), preProc.outBuffer.GetLen() );
							 firstChunk = FALSE;
						}
						else
						{
							InsertNextChunkSpc( preProc.outBuffer.GetTxt(), preProc.outBuffer.GetLen() );
						}
					}
					break;

				default:
					if ( firstChunk )
					{
						 expParams[ parmIndex ] = InsertFirstChunk( preProc.outBuffer.GetTxt(), preProc.outBuffer.GetLen() );
						 firstChunk = FALSE;
					}
					else
					{
						InsertNextChunkSpc( preProc.outBuffer.GetTxt(), preProc.outBuffer.GetLen() );
					}
			}
			token = preProc.tokenizer.Get();
		}
WhileEnd:
		preProc.fileStack.Pop();
	}
	preProc.macroCallStack.SetBottom( savedMCallStkBottom );
	preProc.fileStack.SetBottom( savedFilStkBottom );
}

//##ModelId=419A04D40184
FSMacroCallParams::~FSMacroCallParams()
{}

//-------------------------------------------------------------------
//   class FSMacroCall implementation
//-------------------------------------------------------------------

//##ModelId=419A04D402C5
int FSMacroCall::GetChar()
{  int c;
	while ( ( c = *( currCharPos++ ) ) == '\0' )
	{
		if ( ( readParam = ( Boolean )!readParam ) != FALSE )
		{
			 // reading actual parameter
			 if ( currBodyElem->argMode == FS_AR_NORMAL )
			 {
				 currCharPos = params.GetNonExpanded( currBodyElem->arg );
				 stringifyChar = FALSE;
			 }
			 else if ( currBodyElem->argMode == FS_AR_EXPAND )
			 {
				currCharPos = params.GetExpanded( currBodyElem->arg );
				stringifyChar = FALSE;
			 }
			 else if ( currBodyElem->argMode == FS_AR_STRING )
			 {
				currCharPos = params.GetNonExpanded( currBodyElem->arg );
				stringifyChar = TRUE;
				charInserted = FALSE;
			 }
          else if ( currBodyElem->argMode == FS_AR_FILNAME )
          {
            currCharPos = preProc.File();
            stringifyChar = FALSE;
          }
          else // currBodyElem->argMode == FS_AR_LINENUM
          {
            currCharPos = preProc.LineString();
            stringifyChar = FALSE;
          }
		}
		// end of reading param, go to the next chunk
		else if ( macro.Body().MoreChunks() )
		{
			currBodyElem = macro.Body().NextChunk();
			currCharPos = currBodyElem->txt;
			stringifyChar = FALSE;
		}
		else
		{
			// provision for calling more than once at end of macro body
			currCharPos--;
			readParam = ( Boolean )!readParam;
			c = EOF;
			break;
		}
	}
	if ( stringifyChar ) // special processing of some chars
	{
		if ( charInserted ) // char was inserted in the previous call
		{
			charInserted = FALSE;
		}
		else
		{
			switch( c )
			{
				case '\\': case '"':
				  c = '\\';
				  currCharPos--;
				  charInserted = TRUE;
				// default: no action
			}
		}
	}
	return c;
}

//-------------------------------------------------------------------
//   class FSMacroCallStack implementation
//-------------------------------------------------------------------

//##ModelId=419A04D403DD
Boolean FSMacroCallStack::ExpandingNow( FSMacro *pMacro )
{
   // whole stack is examined, regardless of the current setting of bottom
	for ( int currStkElem = 0; currStkElem < top; currStkElem++ )
	{
		if ( &( stack[ currStkElem ]->Macro() ) == pMacro ) return TRUE;
	}
	return FALSE;
}

//##ModelId=419A04D403D3
void FSMacroCallStack::Push( FSMacro &macro )
{
#ifdef PPDEBUG
   cout << preProc.File() << '(' << preProc.Line() << ") ";
   cout << "MS PUSH MACRO top " << ( top + 1 ) << '\n';
#endif
	// the following statement can involve stack change and also
	// recursive calls to Push during mcall params parsing
	FSMacroCall *pCall = new FSMacroCall( macro, preProc );
	if ( top < FS_MAX_EXPNS_LEVEL )
	{
		stack[ top++ ] = pCall;
		// tokenizer lookahead, which points after the macro call
		// is saved to be used after macro is fully processed
		pCall->savedTokenizerLookahead = preProc.tokenizer.nextChar;
		preProc.tokenizer.nextChar = preProc.inputBuffer.GetChar();
	}
	else
	{
		 delete pCall;
		 FSAnzError( E_PP_MACRO_INCL_OVERFLOW );
	}
}

//##ModelId=419A04D403E8
void FSMacroCallStack::Pop()
{
#ifdef PPDEBUG
   cout << preProc.File() << '(' << preProc.Line() << ") ";
   cout << "MS POP top " << ( top - 1 ) << '\n';
#endif
	preProc.tokenizer.nextChar = stack[ top - 1 ]->savedTokenizerLookahead;
	delete stack[ --top ];
}

//##ModelId=419A04D5001D
FSMacroCallStack::~FSMacroCallStack()
{
	bottom = 0;
   while( IsNotEmpty() )
          Pop();
}

//-------------------------------------------------------------------
//   class FSIfStack implementation
//-------------------------------------------------------------------

//##ModelId=419A04D5014A
void FSIfStack::Push( Boolean ignore, Boolean skipText, unsigned strtLine )
{
#ifdef PPDEBUG
   cout << preProc->File() << '(' << preProc->Line() << ") ";
   cout << "IS PUSH IF top " << ( top + 1 ) << "ign " << ignore << " skip " << skipText << '\n';
#endif
	if ( top < FS_MAX_EXPNS_LEVEL )
	{
		stack[ top++ ] = 	new FSIf( ignore, skipText, strtLine );
	}
	else
	{
       // if stack overflow
       // DA-C anlz. has no such error
		 FSAnzError( E_PP_MACRO_INCL_OVERFLOW );
	}
}

//##ModelId=419A04D50168
void FSIfStack::Clear()
{
   while ( IsNotEmpty() ) Pop();
#ifdef PPDEBUG
   cout << preProc->File() << '(' << preProc->Line() << ") ";
   cout << "IS CLEAR top " << top << '\n';
#endif
}

