//----------------------------------------------------------------------------
//  Description :
//      preprocessor input & output buffers and
//      tokenizer implementation
//----------------------------------------------------------------------------
//  Changes:
//  $Log__$
//----------------------------------------------------------------------------

#ifndef FSTOKEN_H
#define FSTOKEN_H

#ifndef FSSTACKS_H
   #error fsstacks.h required!
#endif

static const int FS_MAX_OUT_BUFFER = 255;

//##ModelId=419A04D201B4
class FSInputBuffer
{
   public:

	//##ModelId=419A04D201BD
      int GetChar();
	//##ModelId=419A04D201C8
      FSInputBuffer( FSFileStack &fileStck, FSMacroCallStack &mcroCallStck )
         : fileStack( fileStck ),
           macroCallStack( mcroCallStck )
      {}

   private:

	//##ModelId=419A04D20205
      FSFileStack &fileStack;
	//##ModelId=419A04D2020F
      FSMacroCallStack &macroCallStack;
};

//-------------------------------------------------------------------
//   class FSInputBuffer implementation
//-------------------------------------------------------------------
//##ModelId=419A04D201BD
inline int FSInputBuffer::GetChar()
{
   return ( macroCallStack.IsNotEmpty() )
            ? macroCallStack.Top().GetChar()
            : ( fileStack.IsNotEmpty() )
              ? fileStack.Top().GetChar()
              : EOF;
}

//##ModelId=419A04D2022C
class FSOutBuffer
{
   public:

	//##ModelId=419A04D20236
      static char *yytext;
	//##ModelId=419A04D20240
      static int yyleng;

	//##ModelId=419A04D2024A
      static char *GetTxt() { return yytext; }
	//##ModelId=419A04D20254
      static int GetLen() { return yyleng; }
	//##ModelId=419A04D20256
      static void Put( char c ) { yytext[ yyleng ] = c; yyleng += ( yyleng < FS_MAX_OUT_BUFFER ); }
	//##ModelId=419A04D20260
      static void RedirectToString( char *str ) { yytext = str; yyleng = strlen( yytext ); }
	//##ModelId=419A04D2026A
      static void RedirectToInternalBuffer() { yytext = outBuffer; yyleng = strlen( yytext ); }
	//##ModelId=419A04D20273
      static void Clear() { yyleng = 0; }
	//##ModelId=419A04D20275
      static void SetTerminator() { yytext[ yyleng ] = '\0'; }
	//##ModelId=419A04D2027D
      FSOutBuffer() { yytext = outBuffer; yyleng = 0; }

   private:

	//##ModelId=419A04D2027E
      static char outBuffer[ FS_MAX_OUT_BUFFER + 1 ];
};

//##ModelId=419A04D202AE
class FSTokenizer
{
	public:

	//##ModelId=419A04D202B8
		int nextChar;
	//##ModelId=419A04D202CD
		Boolean processBackslashInString;
	//##ModelId=419A04D202D7
      Boolean recognizeAsmComments;
	//##ModelId=419A04D202DB
		int Get();
	//##ModelId=419A04D202E0
		FSTokenizer( FSPreProc &pProc, FSInputBuffer &inBuf, FSOutBuffer &outBuf )
			: nextChar( ' ' ),
			  processBackslashInString( TRUE ),
           recognizeAsmComments( FALSE ),
			  preProc( pProc ),
			  inputBuffer( inBuf ),
			  outBuffer( outBuf )
		{}

	private:

	//##ModelId=419A04D202F5
		FSPreProc &preProc;
	//##ModelId=419A04D202FF
		FSInputBuffer &inputBuffer;
	//##ModelId=419A04D20313
		FSOutBuffer &outBuffer;
};

#else
   #error Already included!
#endif
