//----------------------------------------------------------------------------
//  Description :
//      preprocessor file, macro & if stacks,
//      macro table interface
//----------------------------------------------------------------------------
//  Changes:
//  $Log__$
//----------------------------------------------------------------------------
#ifndef FSSTACKS_H
#define FSSTACKS_H

#if !defined( _INC_FSTREAM ) || !defined( _INC_STRING ) \
    || !defined( FSERROR_H ) || !defined( FSBOOL_H ) \
    || ( defined( PPDEBUG ) && ( !defined( _INC_IOSTREAM ) || !defined( FSPPROC_H )))
   #error fstream.h, string.h, fserror.h, fsbool.h and \
   iostream.h and fspproc.h ( if DEBUG ) required!
#endif

static const FS_MAX_ARGS = 20;
static const FS_MAX_EXPNS_LEVEL = 50;
static const FS_MAX_INCLUDE_LEVEL = 50;
static const FS_MAX_IF_NESTING_LEVEL = 200;
static const FS_HASH_DIM = 53;
static const FS_PARAM_STACK_DIM = 5000;
static const FS_MAX_FILE_NAME = 80;
static const FS_MAX_PARAM_LEN = 2000;
static const FS_MAX_MBODY_CHUNK_LEN = 80;

//##ModelId=419A04D2031C
enum FSStackElemType { 
	//##ModelId=419A04D20331
	FS_STK_FILE, 
	//##ModelId=419A04D2033A
	FS_STK_PARAM 
};

//##ModelId=419A04D2034E
class FSFileStackElem
{
   public:

	//##ModelId=419A04D20359
		const FSStackElemType type;
	//##ModelId=419A04D2036C
      int savedTokenizerLookahead;
	//##ModelId=419A04D20376
      virtual int GetChar() = 0;
	//##ModelId=419A04D20378
      virtual char *Name() { return 0; }
	//##ModelId=419A04D20381
      virtual unsigned Line() { return 0; }
	//##ModelId=419A04D2038A
      FSFileStackElem( FSStackElemType typ )
         : type( typ )
      {}
	//##ModelId=419A04D20394
      virtual ~FSFileStackElem() {}
};

class FSPreProc;

//##ModelId=419A04D203A8
class FSFile: public FSFileStackElem
{
	public:

	//##ModelId=419A04D30056
		enum OpenStatus { 
		//##ModelId=419A04D30061
			SUCCESS, 
		//##ModelId=419A04D3006A
			ERROR
	};
	//##ModelId=419A04D203BC
		int GetChar();

	//##ModelId=419A04D203BD
		virtual char *Name() { return name; }
	//##ModelId=419A04D203C7
		virtual unsigned Line() { return line; }
	//##ModelId=419A04D203D1
		OpenStatus Status() { return ( OpenStatus )( file.fail() ); }

	//##ModelId=419A04D203D2
		FSFile( FSPreProc &pProc, char *fileName = 0 );

	//##ModelId=419A04D203E4
		virtual ~FSFile();

   private:

	//##ModelId=419A04D203E7
      FSPreProc &preProc;
	//##ModelId=419A04D30008
      char name[ FS_MAX_FILE_NAME + 1 ];
	//##ModelId=419A04D3001A
      int lookAheadChar;
	//##ModelId=419A04D30024
      int savedIfStackBottom;
	//##ModelId=419A04D30039
      unsigned line;
	//##ModelId=419A04D30043
      ifstream file;
};

//##ModelId=419A04D3007E
class FSParam: public FSFileStackElem
{
   public:

	//##ModelId=419A04D30093
      int GetChar()
      {
         int temp = *( currPos++ );
         if ( temp != 0 )
            return temp;
         else
         {
            currPos--;
            // multiple calls at end of string are allowed
            // \n before EOF not applicable in this situation
            return EOF;
         }
      }

	//##ModelId=419A04D30094
      FSParam( char *str )
         : FSFileStackElem( FS_STK_PARAM ),
           currPos( str )
      {}

	//##ModelId=419A04D3009E
      virtual ~FSParam() {}

	private:

	//##ModelId=419A04D300A7
		char *currPos;
};

//##ModelId=419A04D300CF
class FSFileStack
{
	public:

	//##ModelId=419A04D300D9
		Boolean IsNotEmpty() { return (Boolean) ( top != bottom ); }

	//##ModelId=419A04D300E3
		void PushParam( char *str );

	//##ModelId=419A04D300ED
		void PushFile( char *fileName = 0 );

	//##ModelId=419A04D300F7
		FSFileStackElem &Top() { return *( stack[ top - 1 ] ) ; }

	//##ModelId=419A04D30101
		void Pop();

	//##ModelId=419A04D30102
		int SetBottom()
      {
#ifdef PPDEBUG
         cout << preProc.File() << '(' << preProc.Line() << ") ";
         cout << "FS SET BOTTOM old " << bottom << " new " << top << '\n';
#endif
         int temp = bottom; bottom = top; return temp;
      }

	//##ModelId=419A04D3010B
		int SetBottom( int newBottom )
      {
#ifdef PPDEBUG
         cout << preProc.File() << '(' << preProc.Line() << ") ";
         cout << "FS SET BOTTOM old " << bottom << " new " << newBottom << '\n';
#endif
         int temp = bottom; bottom = newBottom; return temp;
      }

	//##ModelId=419A04D30115
		FSFileStack( FSPreProc &pProc )
			: preProc( pProc ),
			  top( 0 ),
			  bottom( 0 )
		{}

	//##ModelId=419A04D30117
		~FSFileStack();

	private:

	//##ModelId=419A04D30120
		FSPreProc &preProc;
	//##ModelId=419A04D30129
		int top;
	//##ModelId=419A04D3013D
		int bottom;
	//##ModelId=419A04D30152
		FSFileStackElem *stack[ FS_MAX_INCLUDE_LEVEL ];
};


//##ModelId=419A04D30183
class FSMacroArgs
{
   public:

	//##ModelId=419A04D3018D
      int LookUp( char *argName );
	//##ModelId=419A04D30197
      Boolean &Parenth() { return parenth; }
	//##ModelId=419A04D30198
      int &Number() { return number; }
	//##ModelId=419A04D301A1
		FSMacroArgs( FSPreProc &preProc, int &lastTokenRead ); // parsira definiciju makroa iza imena do tela
	//##ModelId=419A04D301AC
      FSMacroArgs(); // for predefined macro without args
	//##ModelId=419A04D301B5
      ~FSMacroArgs();

	private:

	//##ModelId=419A04D301B7
      Boolean parenth;
	//##ModelId=419A04D301BF
      int number;
	//##ModelId=419A04D301D3
      char *args[ FS_MAX_ARGS + 1 ];

	//##ModelId=419A04D301DD
      void Insert( char *argName );
};

//##ModelId=419A04D301FB
enum FSArgumentMode { 
	//##ModelId=419A04D30206
	FS_AR_NORMAL, 
	//##ModelId=419A04D3020F
	FS_AR_STRING,
	//##ModelId=419A04D30219
	FS_AR_EXPAND, 
	//##ModelId=419A04D30223
	FS_AR_FILNAME, 
	//##ModelId=419A04D3022D
	FS_AR_LINENUM 
};

//##ModelId=419A04D3025F
class FSMacroBody
{
	public:

	//##ModelId=419A04D302CD
		class FSMacroBodyElem
		{
			public:

		//##ModelId=419A04D302D7
				char *txt;
		//##ModelId=419A04D302E1
				int arg;
		//##ModelId=419A04D302EC
				FSArgumentMode argMode;
		//##ModelId=419A04D302F8
				FSMacroBodyElem *next;

				// appends a string str onto the txt component; in case there is not
				// enough space left, allocates a new Elem, chains it to the list,
				// updates the current ptr, and the rest of string str inserts in the new txt
		//##ModelId=419A04D302FF
				static void Insert( FSMacroBodyElem *&current, char *str, int len );
		//##ModelId=419A04D3031D
				static void Insert( FSMacroBodyElem *&current, int arg, FSArgumentMode mode );
		//##ModelId=419A04D30331
				FSMacroBodyElem()
					: txt( new char[ FS_MAX_MBODY_CHUNK_LEN + 1 ] ),
					  arg( 0 ),
                 argMode( FS_AR_NORMAL ),
					  next( 0 ),
					  txtEnd( txt )
				{
					*txt = '\0';
				}

		//##ModelId=419A04D30332
				~FSMacroBodyElem() { delete [] txt; }

			private:

		//##ModelId=419A04D3033B
				char *txtEnd;
		};

	//##ModelId=419A04D30269
		FSMacroBodyElem *FirstChunk() { return current = head; };
	//##ModelId=419A04D30273
		Boolean MoreChunks() { return ( Boolean ) ( current->next != 0 ); }
	//##ModelId=419A04D30274
		FSMacroBodyElem *NextChunk() { return current = current->next; }

	//##ModelId=419A04D3027D
		FSMacroBody( FSPreProc &preProc, FSMacroArgs &args, int &lastTokenRead ); // parsira telo makroa
	//##ModelId=419A04D30289
      FSMacroBody(); // for predefined macros: null body
	//##ModelId=419A04D30291
      ~FSMacroBody();

	private:

	//##ModelId=419A04D30293
      FSMacroBodyElem *head;
	//##ModelId=419A04D302A6
      FSMacroBodyElem *current;
};

//##ModelId=419A04D30378
class FSMacro
{
   public:

	//##ModelId=419A04D3038C
		char *Name() { return name; }
	//##ModelId=419A04D3038D
      Boolean &Redefinable() { return redefinable; }
	//##ModelId=419A04D30396
		FSMacroArgs &Args() { return args; }
	//##ModelId=419A04D30397
		FSMacroBody &Body() { return body; }
	//##ModelId=419A04D303A0
		FSMacro *&Next() { return next; }
	//##ModelId=419A04D303A1
		FSMacro( char *macroName, FSPreProc &preProc, int &lastTokenRead, Boolean redefinabl = TRUE )
			: name( strcpy( new char[ strlen( macroName ) + 1 ], macroName ) ),
           redefinable( redefinabl ),
			  args( preProc, lastTokenRead ),
			  body( preProc, args, lastTokenRead ),
           next( 0 )
      {
#ifdef PPDEBUG
         cout << preProc.File() << '(' << preProc.Line() << ") ";
         cout << "MACRODEF "<< Name() << ' ' << Args().Number() << " args\n";
#endif
      }

      // for predefined macros: creates empty non redefinable macro
	//##ModelId=419A04D303B5
		FSMacro( char *macroName, Boolean redefinabl = FALSE )
			: name( strcpy( new char[ strlen( macroName ) + 1 ], macroName ) ),
           redefinable( redefinabl ),
           next( 0 )
      {
#ifdef PPDEBUG
         //cout << preProc.File() << '(' << preProc.Line() << ") ";
         cout << "BUILTINMACRODEF "<< Name() << ' ' << Args().Number() << " args\n";
#endif
      }

	//##ModelId=419A04D303C0
      ~FSMacro() { delete [] name; }

	private:

	//##ModelId=419A04D303C8
      char *name;
	//##ModelId=419A04D303D3
      Boolean redefinable; // __STDC__ etc. are not
	//##ModelId=419A04D303DD
		FSMacroArgs args;
	//##ModelId=419A04D40009
		FSMacroBody body;
	//##ModelId=419A04D4001F
      FSMacro *next;
};

//##ModelId=419A04D4006C
class FSMacroTable
{
	public:

	//##ModelId=419A04D40076
		FSMacro *LookUp( char *name ) { return *InternLookUp( name ); }

	//##ModelId=419A04D40081
      void Insert( FSMacro *newMacro ) // for predefined macros
		{
			FSMacro **ppMacro = InternLookUp( newMacro->Name() );
			if ( *ppMacro == 0 )
			{
				InternInsert( newMacro, ppMacro );
			}
			else
			{
            // !!! provera jednakosti redefinicije i eventualna prijava greske
            if ( ( *ppMacro )->Redefinable() )
            {
				   InternInsert( newMacro, ppMacro );
				   InternRemove( &( newMacro->Next() ) );
            }
            else
            {
               // else error : can not redefine macro; DA-C not reporting
               delete newMacro;
            }
			}
		}

	//##ModelId=419A04D4008B
		void Insert( char *name, int &lastTokenRead ) // destroys name value !!!
		{
			FSMacro *newMacro = new FSMacro( name, preProc, lastTokenRead );
			FSMacro **ppMacro = InternLookUp( newMacro->Name() );
			if ( *ppMacro == 0 )
			{
				InternInsert( newMacro, ppMacro );
			}
			else
			{  // !!!provera jednakosti redefinicije i event. prijava greske
            if ( ( *ppMacro )->Redefinable() )
            {
				   InternInsert( newMacro, ppMacro );
				   InternRemove( &( newMacro->Next() ) );
            }
            else
            {
               // else error : can not redefine macro; DA-C not reporting
               delete newMacro;
            }
			}
		}

	//##ModelId=419A04D4009E
		void Remove( char *name )
		{
			FSMacro **ppMacro = InternLookUp( name );
			if ( *ppMacro != 0 ) // #undef of undefined macro is not an error
         {
            if ( ( *ppMacro )->Redefinable() )
			   {
               InternRemove ( ppMacro );
            }
            // else error: can not #undef macro; DA-C not reporting
         }
		}

	//##ModelId=419A04D400A0
		FSMacroTable( FSPreProc &pProc );
	//##ModelId=419A04D400A9
		~FSMacroTable();

	private:

	//##ModelId=419A04D400B3
		FSPreProc &preProc;

	//##ModelId=419A04D400BD
		FSMacro *table[ FS_HASH_DIM + 1 ]; // last element is a list of #undefined macros

	//##ModelId=419A04D400C6
		unsigned HashValue( char *name );
		// returns a pointer to macro 'name' if it is present in the table, or a
		// pointer to 'next' field in the last elem of the hash list where the
		// macro should be placed
	//##ModelId=419A04D400DA
		FSMacro **InternLookUp( char *name );

		// adds a macro to a hash chain
	//##ModelId=419A04D400E4
		void InternInsert( FSMacro *pMacro, FSMacro **ppWhere )
		{
			pMacro->Next() = *ppWhere;
			*ppWhere = pMacro;
		}
		// removes a macro pointed to by ppMacro from hash chain
		// *ppMacro must be != 0
	//##ModelId=419A04D400F8
		void InternRemove( FSMacro **ppMacro )
		{
			FSMacro *toDel = *ppMacro;
			*ppMacro = toDel->Next();
			InternInsert( toDel, table + FS_HASH_DIM );
		}
};

class FSMacroCallStack;

//##ModelId=419A04D4015C
class FSMacroCallParams
{
	public:

	//##ModelId=419A04D40166
		char *GetNonExpanded( int paramIndex ) { return nonExpParams[ paramIndex ]; }
	//##ModelId=419A04D40170
		char *GetExpanded( int paramIndex ) { return expParams [ paramIndex ]; }

	//##ModelId=419A04D4017A
		FSMacroCallParams( FSMacro &mcro, FSPreProc &preProc );

	//##ModelId=419A04D40184
		~FSMacroCallParams();

	private:

	//##ModelId=419A04D4022F
		enum Status { 
		//##ModelId=419A04D40239
			EXPECT_ARG, 
		//##ModelId=419A04D40243
			EXPECT_ARG_OR_SEPARATOR, 
		//##ModelId=419A04D4024D
			END_LIST 
		};
	//##ModelId=419A04D40186
		FSMacro &macro;
	//##ModelId=419A04D40199
		FSPreProc &preProc;
	//##ModelId=419A04D401A2
		char *nonExpParams[ FS_MAX_ARGS + 1 ]; // first param in nonExpParams[1]
	//##ModelId=419A04D401AC
		char *expParams[ FS_MAX_ARGS + 1 ];
	//##ModelId=419A04D401C0
		char paramSpace[ FS_MAX_PARAM_LEN + 1 ];
	//##ModelId=419A04D401CA
		char *paramSpcFree;

	//##ModelId=419A04D401D4
		void ProcessParamChunk( int &numParams, Status &status );

	//##ModelId=419A04D401E8
		char *InsertFirstChunk( char *string, int len ) //  space before
		{
			char *result = "";
			if ( paramSpcFree + len < paramSpace + FS_MAX_PARAM_LEN )
			{
				result = strcpy( paramSpcFree, string );
				paramSpcFree += len + 1;
			}
			else
         {
            // param space overflow
             paramSpcFree = paramSpace + FS_MAX_PARAM_LEN;
				 FSAnzError( E_PP_MACRO_INCL_OVERFLOW );
         }
			return result;
		}

	//##ModelId=419A04D401FC
		void InsertNextChunkSpc( char *string, int len )
		{
			if ( paramSpcFree + len < paramSpace + FS_MAX_PARAM_LEN )
			{
				*( paramSpcFree - 1) = ' ';
				strcpy( paramSpcFree, string );
				paramSpcFree += len + 1;
			}
			else
			{
            // param space overflow
             paramSpcFree = paramSpace + FS_MAX_PARAM_LEN;
				 FSAnzError( E_PP_MACRO_INCL_OVERFLOW );
			}
		}
};

//##ModelId=419A04D402B1
class FSMacroCall
{
	public:

	//##ModelId=419A04D402BB
		int savedTokenizerLookahead;

	//##ModelId=419A04D402C5
		int GetChar();

	//##ModelId=419A04D402CF
		FSMacro &Macro() { return macro; }

	//##ModelId=419A04D402D0
		FSMacroCall( FSMacro &mcro, FSPreProc &prProc )
			: preProc( prProc ),
           macro( mcro ),
			  params( mcro, preProc ),
			  currBodyElem( mcro.Body().FirstChunk() ),
			  currCharPos( currBodyElem->txt ),
			  readParam( FALSE ),
			  stringifyChar( FALSE ),
			  charInserted( FALSE )
		{
#ifdef PPDEBUG
// supress 'out of line inline' warning
#pragma warn -inl
         cout << preProc.File() << '(' << preProc.Line() << ") ";
         cout << "MACRO "<< macro.Name() << ' ' << macro.Args().Number() << " args, redefinable: " << macro.Redefinable() << "\nparams:";
			for( int i = 1; i <= macro.Args().Number(); i++ )
            cout << "    num " << i << ":" << params.GetNonExpanded( i ) << ":" << params.GetExpanded( i );
			cout << "\nbody";
			while ( 1 )
			{
            cout << "     :" << currBodyElem->txt << ":, (" <<
						 ( ( currBodyElem->argMode == FS_AR_NORMAL )
							  ? "norm"
							  : ( currBodyElem->argMode == FS_AR_STRING )
								  ? "str"
								  : ( currBodyElem->argMode == FS_AR_FILNAME )
                             ? "file"
                             :( currBodyElem->argMode == FS_AR_LINENUM )
                               ? "line"
                               : "exp"
                   )
                << ')' << currBodyElem->arg;

				if ( macro.Body().MoreChunks() == FALSE ) break;
				currBodyElem = macro.Body().NextChunk();
			}
			currBodyElem = macro.Body().FirstChunk();
         cout << '\n';
#pragma warn .inl
#endif
		}

	//##ModelId=419A04D402DA
      ~FSMacroCall() {}

	private:

	//##ModelId=419A04D402E4
      FSPreProc &preProc;
	//##ModelId=419A04D402EE
		FSMacro &macro;
	//##ModelId=419A04D40302
		FSMacroCallParams params;
	//##ModelId=419A04D4030C
		FSMacroBody::FSMacroBodyElem *currBodyElem;
	//##ModelId=419A04D40315
		char *currCharPos;
	//##ModelId=419A04D4032A
		Boolean readParam;
	//##ModelId=419A04D40334
		Boolean stringifyChar;
	//##ModelId=419A04D40348
		Boolean charInserted;
};

//##ModelId=419A04D403BF
class FSMacroCallStack
{
	public:

	//##ModelId=419A04D403C9
		Boolean IsNotEmpty() { return (Boolean) ( top != bottom ); }

	//##ModelId=419A04D403D3
		void Push( FSMacro &macro );

	//##ModelId=419A04D403DD
		Boolean ExpandingNow( FSMacro *pMacro );

	//##ModelId=419A04D403E7
		FSMacroCall &Top() { return *( stack[ top - 1 ] ); } // should not be called for empty stack

	//##ModelId=419A04D403E8
		void Pop(); // should not be called for empty stack

	//##ModelId=419A04D50009
		int SetBottom()
      {
#ifdef PPDEBUG
         cout << preProc.File() << '(' << preProc.Line() << ") ";
         cout << "MS SET BOTTOM old " << bottom << " new " << top << '\n';
#endif
         int temp = bottom; bottom = top; return temp;
      }

	//##ModelId=419A04D5000A
		int SetBottom( int newBottom )
      {
#ifdef PPDEBUG
         cout << preProc.File() << '(' << preProc.Line() << ") ";
         cout << "MS SET BOTTOM old " << bottom << " new " << newBottom << '\n';
#endif
         int temp = bottom; bottom = newBottom; return temp;
      }

	//##ModelId=419A04D50014
		FSMacroCallStack( FSPreProc &pProc )
			: preProc( pProc ),
			  top( 0 ),
			  bottom( 0 )
		{}

	//##ModelId=419A04D5001D
		~FSMacroCallStack();


	private:

	//##ModelId=419A04D5001F
		FSPreProc &preProc;
	//##ModelId=419A04D50032
		FSMacroCall *stack[ FS_MAX_EXPNS_LEVEL ];
	//##ModelId=419A04D5003B
		int top;
	//##ModelId=419A04D50045
		int bottom;
};

//##ModelId=419A04D50081
class FSIf
{
	public:

	//##ModelId=419A04D5008C
      Boolean ignore;
	//##ModelId=419A04D50096
		Boolean elsePart;
	//##ModelId=419A04D500AA
		Boolean skipText;
	//##ModelId=419A04D500B4
      unsigned startLine;

	//##ModelId=419A04D500BD
		FSIf( Boolean ignre, Boolean skipTxt, unsigned strtLine )
			: ignore( ignre ),
           elsePart( FALSE ),
			  skipText ( skipTxt ),
           startLine( strtLine )
		{}
};

//##ModelId=419A04D50136
class FSIfStack
{
	public:

	//##ModelId=419A04D50140
		Boolean IsNotEmpty() { return (Boolean) ( top != bottom ); }

	//##ModelId=419A04D5014A
		void Push( Boolean ignore, Boolean skipText, unsigned strtLine );

	//##ModelId=419A04D50156
		FSIf &Top() { return *( stack[ top - 1 ] ); } // should not be called for empty stack

	//##ModelId=419A04D5015E
		void Pop()  // should not be called for empty stack
      {
#ifdef PPDEBUG
         cout << preProc->File() << '(' << preProc->Line() << ") ";
         cout << "IS POP top " << (top-1) << '\n';
#endif
         delete stack[ --top ];
      }

	//##ModelId=419A04D50168
      void Clear();

	//##ModelId=419A04D50169
      int SetBottom()
      {
#ifdef PPDEBUG
         cout << preProc->File() << '(' << preProc->Line() << ") ";
         cout << "IS SET BOTTOM old " << bottom << " new " << top << '\n';
#endif
         int temp = bottom; bottom = top; return temp;
      }

	//##ModelId=419A04D50172
		int SetBottom( int newBottom )
      {
#ifdef PPDEBUG
         cout << preProc->File() << '(' << preProc->Line() << ") ";
         cout << "IS SET BOTTOM old " << bottom << " new " << newBottom << '\n';
#endif
         int temp = bottom; bottom = newBottom; return temp;
      }

	//##ModelId=419A04D50174
      FSIfStack()
         : top( 0 ),
           bottom( 0 )
      {}

	//##ModelId=419A04D5017C
      ~FSIfStack() { SetBottom( 0 ); Clear(); }

	private:

	//##ModelId=419A04D5017E
		FSIf *stack[ FS_MAX_IF_NESTING_LEVEL ];
	//##ModelId=419A04D50190
		int top;
	//##ModelId=419A04D5019A
      int bottom;
};

#else
   #error Already included!
#endif
